from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
for zip_name in ["ui","materials","operators","presets","properties","pyfluid","resources","source"]:
    download_zip(xz_url + zip_name +".zip",down_path)

from .translation.translate import GetTranslationDict

bl_info = {
    "name" : "FLIP Fluids",
    "description": "A FLIP Fluid Simulation Tool for Blender (v1.8.1 Release 16-JUL-2024)",
    "author" : "Ryan Guy & Dennis Fassbaender",
    "version" : (1, 8, 1),
    "blender" : (3, 1, 0),
    "location" : "Properties > Physics > FLIP Fluid",
    "warning" : "",
    "wiki_url" : "",
    "doc_url" : "",
    "category" : "Animation"
}

if "bpy" in locals():
    import importlib
    reloadable_modules = [
        'filesystem',
        'utils',
        'objects',
        'materials',
        'properties',
        'operators',
        'ui',
        'presets',
        'export',
        'bake',
        'render',
        'exit_handler'
    ]
    for module_name in reloadable_modules:
        if module_name in locals():
            importlib.reload(locals()[module_name])

import bpy, atexit, shutil, os
from bpy.props import (
        PointerProperty,
        StringProperty
        )

from . import (
        filesystem,
        utils,
        objects,
        materials,
        properties,
        operators,
        ui,
        presets,
        export,
        bake,
        render,
        exit_handler
        )

from .utils import installation_utils
from .utils import version_compatibility_utils as vcu


@bpy.app.handlers.persistent
def scene_update_post(scene):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    installation_utils.scene_update_post(scene)
    if installation_utils.is_addon_active():
        if not render.is_rendering():
            # We don't want to update these while rendering to prevent
            # odd behaviour in the depsgraph
            properties.scene_update_post(scene)
            materials.scene_update_post(scene)
        render.scene_update_post(scene)


@bpy.app.handlers.persistent
def render_init(scene):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    render.render_init(scene)


@bpy.app.handlers.persistent
def render_complete(scene):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    render.render_complete(scene)


@bpy.app.handlers.persistent
def render_cancel(scene):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    render.render_cancel(scene)


@bpy.app.handlers.persistent
def frame_change_pre(scene, depsgraph=None):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return


@bpy.app.handlers.persistent
def frame_change_post(scene, depsgraph=None):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    properties.frame_change_post(scene, depsgraph)
    render.frame_change_post(scene, depsgraph)


@bpy.app.handlers.persistent
def render_pre(scene, depsgraph=None):
    if scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    render.render_pre(scene)


@bpy.app.handlers.persistent
def load_pre(nonedata):
    if bpy.context.scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    properties.load_pre()


@bpy.app.handlers.persistent
def load_post(nonedata):
    if bpy.context.scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    if vcu.is_blender_28() and not vcu.is_blender_281():
        print("FLIP FLUIDS WARNING: Blender 2.80 contains bugs that can cause frequent crashes during render, Alembic export, and rigid/cloth simulation baking. Blender version 2.81 or higher is recommended.")

    installation_utils.load_post()
    materials.load_post()
    properties.load_post()
    presets.load_post()
    exit_handler.load_post()
    

@bpy.app.handlers.persistent
def save_pre(nonedata):
    if bpy.context.scene.flip_fluid.is_addon_disabled_in_blend_file():
        return

    properties.save_pre()


@bpy.app.handlers.persistent
def save_post(nonedata):
    if bpy.context.scene.flip_fluid.is_addon_disabled_in_blend_file():
        return
        
    properties.save_post()
    exit_handler.save_post()


def on_exit():
    exit_handler.on_exit()


class FlipFluidCompleteInstallation(bpy.types.Operator):
    bl_idname = "flip_fluid_operators.complete_installation"
    bl_label = "Complete Installation"
    bl_description = ("Click to complete the installation of the FLIP Fluids addon. Alternatively, restarting Blender or re-loading the Blend file will also complete the installation process")


    @classmethod
    def poll(cls, context):
        return True


    def execute(self, context):
        load_post(None)
        return {'FINISHED'}


def register():
    objects.register()
    materials.register()
    properties.register()
    operators.register()
    ui.register()
    presets.register()

    if vcu.is_blender_28():
        bpy.app.handlers.depsgraph_update_post.append(scene_update_post)
    else:
        bpy.app.handlers.scene_update_post.append(scene_update_post)

    bpy.app.handlers.render_init.append(render_init)
    bpy.app.handlers.render_complete.append(render_complete)
    bpy.app.handlers.render_cancel.append(render_cancel)
    bpy.app.handlers.frame_change_pre.append(frame_change_pre)
    bpy.app.handlers.frame_change_post.append(frame_change_post)
    bpy.app.handlers.render_pre.append(render_pre)
    bpy.app.handlers.load_pre.append(load_pre)
    bpy.app.handlers.load_post.append(load_post)
    bpy.app.handlers.save_pre.append(save_pre)
    bpy.app.handlers.save_post.append(save_post)
    atexit.register(on_exit)

    bpy.utils.register_class(FlipFluidCompleteInstallation)
    bpy.app.translations.register(__package__, GetTranslationDict())




def unregister():
    objects.unregister()
    materials.unregister()
    properties.unregister()
    operators.unregister()
    ui.unregister()
    presets.unregister()

    if vcu.is_blender_28():
        bpy.app.handlers.depsgraph_update_post.remove(scene_update_post)
    else:
        bpy.app.handlers.scene_update_post.remove(scene_update_post)

    bpy.app.handlers.render_init.remove(render_init)
    bpy.app.handlers.render_complete.remove(render_complete)
    bpy.app.handlers.render_cancel.remove(render_cancel)
    bpy.app.handlers.frame_change_pre.remove(frame_change_pre)
    bpy.app.handlers.frame_change_post.remove(frame_change_post)
    bpy.app.handlers.render_pre.remove(render_pre)
    bpy.app.handlers.load_pre.remove(load_pre)
    bpy.app.handlers.load_post.remove(load_post)
    bpy.app.handlers.save_pre.remove(save_pre)
    bpy.app.handlers.save_post.remove(save_post)
    atexit.unregister(on_exit)

    bpy.utils.unregister_class(FlipFluidCompleteInstallation)
    bpy.app.translations.unregister(__package__)
